// Copyright 2023 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package runtime

import _ "unsafe"

func initSecureMode() {
	// We have already initialized the secureMode bool in sysauxv.
}

func isSecureMode() bool {
	return secureMode
}
