// Copyright 2011 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package types2

import "cmd/compile/internal/syntax"

// A Type represents a type of Go.
// All types implement the Type interface.
type Type = syntax.Type
