// Code generated by mkduff.go; DO NOT EDIT.
// Run go generate from src/runtime to update.
// See mkduff.go for comments.

//go:build ppc64 || ppc64le

#include "textflag.h"

TEXT runtime·duffzero<ABIInternal>(SB), NOSPLIT|NOFRAME, $0-0
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	MOVDU	R0, 8(R20)
	RET

TEXT runtime·duffcopy<ABIInternal>(SB), NOSPLIT|NOFRAME, $0-0
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	MOVDU	8(R20), R5
	MOVDU	R5, 8(R21)
	RET
